//------------------------------------------------------------------------------
// audiorecorder.cpp
//
// Copyright (C) 2013 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "audiorecorder.h"

#include "audio/containers/wavcontainer.h"
#include "audio/audiodevformat.h"
#include "audio/audioerror.h"
#include "audio/filteredaudioinput.h"
#include <QDebug>

class AudioRecorder::PrivData
{
	public:
		FilteredAudioInput* audioInput;
		WAVContainer* outputDevice;
		int lastError;
};

AudioRecorder::AudioRecorder()
{
	d = new PrivData();
	d->audioInput = NULL;
	d->outputDevice = new WAVContainer();
	d->lastError = 0;
}

AudioRecorder::~AudioRecorder()
{
	if (d->audioInput != NULL)
	{
		delete d->audioInput;
	}
	if (d->outputDevice != NULL)
	{
		delete d->outputDevice;
	}
	delete d;
}

bool AudioRecorder::isRecording() const
{
	if (d->audioInput == NULL)
	{
		return false;
	}
	return d->audioInput->isRecording();
}

int AudioRecorder::lastError() const
{
	return d->lastError;
}

void AudioRecorder::onAudioInputStateChange(QAudio::State state)
{
	if (state == QAudio::StoppedState)
	{
		d->outputDevice->close();
	}
	emit stateChanged(state);
}

void AudioRecorder::onInputReadyRead()
{
	QByteArray data = d->audioInput->readPending();
	d->outputDevice->write(data);
}

const QAudioFormat& AudioRecorder::recordingFormat() const
{
	return d->audioInput->devFormat().format();
}

qint64 AudioRecorder::recordingTimeMs() const
{
	if (d->audioInput == NULL)
	{
		return 0;
	}
	return d->audioInput->recordingTimeMs();
}

FilteredAudioInput* AudioRecorder::setupAudioInput()
{
	FilteredAudioInput* pAudioInput = new FilteredAudioInput();
	int setupErrorCode = pAudioInput->setup();
	if (setupErrorCode != 0)
	{
		d->lastError = setupErrorCode;
		return NULL;
	}
	this->connect(pAudioInput, SIGNAL(notify()), SIGNAL(notify()));
	this->connect(pAudioInput, SIGNAL(stateChanged(QAudio::State)),
		SLOT(onAudioInputStateChange(QAudio::State)));
	this->connect(pAudioInput, SIGNAL(readyRead()), SLOT(onInputReadyRead()));

	return pAudioInput;
}

int AudioRecorder::startRecording(QIODevice* outputStream)
{
	if (isRecording())
	{
		// This method is public.
		// Do not allow to start recording twice.
		return AudioError::OperationAlreadyInProgress;
	}
	d->lastError = 0;

	if (d->audioInput != NULL)
	{
		delete d->audioInput;
		d->audioInput = NULL;
	}

	d->audioInput = setupAudioInput();
	if (d->audioInput == NULL)
	{
		return d->lastError;
	}

	// Create and open the output device.
	d->outputDevice->setAudioFormat(recordingFormat());
	d->outputDevice->setIODevice(outputStream);
	if (!d->outputDevice->open(QIODevice::WriteOnly | QIODevice::Truncate))
	{
		qDebug() << "Recording failed";
		qDebug() << "File cannot be opened for writing";
		return AudioError::OutputOpenFailure;
	}

	int startErrorCode = d->audioInput->start();
	if (startErrorCode == 0)
	{
		qDebug() << "Recording started";
	}
	return startErrorCode;
}

void AudioRecorder::stopRecording()
{
	if (isRecording())
	{
		// This method is public.
		// Do not allow to stop recording if it is not running.
		d->audioInput->stop();
	}
}
