//------------------------------------------------------------------------------
// wavcontainer.cpp
//
// Copyright (C) 2011 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "wavcontainer.h"

#include "audio/containers/wavheader.h"

#include <QDebug>
#include <QVariant>
#include <cassert>

WAVContainer::WAVContainer(QIODevice* pIODevice, QObject* pParent)
: QIODevice(pParent)
{
	d.pIODevice = pIODevice;
}

WAVContainer::~WAVContainer()
{
}

void WAVContainer::close()
{
	if (isWritable())
	{
		// Final data size write to ensure that file data is correct.
		writeDataSize();
	}
	QIODevice::close();
}

qint64 WAVContainer::size() const
{
	qint64 size = d.pIODevice->size();
	return size - WAVHeader::HEADER_SIZE;
}

void WAVContainer::setAudioFormat(const QAudioFormat& format)
{
	d.audioFormat = format;
}

void WAVContainer::setIODevice(QIODevice* pDevice)
{
	if (isOpen())
	{
		close();
	}
	d.pIODevice = pDevice;
}

bool WAVContainer::open(QIODevice::OpenMode mode)
{
	assert (d.pIODevice != NULL);
	if (d.pIODevice == NULL)
	{
		return false;
	}

	bool bSuccess = QIODevice::open(mode);
	if (isReadable())
	{
		readWAVHeader(*d.pIODevice);
	}
	return bSuccess;
}

qint64 WAVContainer::readData(char* data, qint64 maxSize)
{
	if (pos() < WAVHeader::HEADER_SIZE)
	{
		seek(WAVHeader::HEADER_SIZE);
	}
	
	return d.pIODevice->read(data, maxSize);
}

qint64 WAVContainer::readDataSize(QIODevice& dev)
{
	//QDataStream stream(&dev);
	//
	//// Store the current device position to maintain integrity.
	//qint64 oldPos = dev.pos();
	//
	//// Go to the position where the size is stored and read it.
	//if (!dev.seek(DATACHUNK_SIZE_POS))
	//{
	//	return false;
	//}
	//qint32 dataSize;
	//
	//stream >> dataSize;
	//
	//// Restore the old position.
	//dev.seek(oldPos);
	//return (qint64)dataSize;
	
	qDebug() << "Reading data size for RIFF WAVE is not supported.";
	return 0;
}

bool WAVContainer::readWAVHeader()
{
	assert(d.pIODevice != NULL);
	assert(isReadable());
	if (d.pIODevice == NULL || !isReadable())
	{
		return false;
	}

	d.audioFormat = readWAVHeader(*d.pIODevice);
	return d.audioFormat.isValid();
}

QAudioFormat WAVContainer::readWAVHeader(QIODevice& dev)
{
	 qDebug() << "Reading RIFF WAVE files is not supported.";
	 return QAudioFormat();
}

bool WAVContainer::reset()
{
	return seek(0);
}

bool WAVContainer::seek(qint64 pos)
{
	if (pos < 0)
	{
		pos = 0;
	}

	if (d.pIODevice->seek(pos))
	{
		QIODevice::seek(pos);
		return true;
	}

	return false;
}

qint64 WAVContainer::writeData(const char* data, qint64 maxSize)
{
	if (pos() < WAVHeader::HEADER_SIZE)
	{
		// This means that we did not write the header, yet.
		writeWAVHeader();
	}
	
	return d.pIODevice->write(data, maxSize);
}

bool WAVContainer::writeDataSize()
{
	assert(d.pIODevice != NULL);
	assert(isWritable());
	
	if (isWritable())
	{
		return writeDataSize(*d.pIODevice, size());
	}
	
	return false;
}

bool WAVContainer::writeWAVHeader()
{
	assert(d.pIODevice != NULL);
	assert(isWritable());
	
	if (isWritable())
	{
		return writeWAVHeader(*d.pIODevice, d.audioFormat);
	}
	
	return false;
}

bool WAVContainer::writeDataSize(QIODevice& dev, qint64 dataSize)
{
	const int CHUNK_SIZE_POS = 4;
	
	// Store the current device position to maintain integrity.
	qint64 oldPos = dev.pos();

	QDataStream stream(&dev);
	stream.setByteOrder(QDataStream::LittleEndian);
	
	if (!dev.seek(CHUNK_SIZE_POS))
	{
		dev.seek(oldPos);
		return false;
	}
	stream << (qint32)(WAVHeader::HEADER_SIZE - 8 + dataSize);
	
	if (!dev.seek(WAVHeader::DATACHUNK_SIZE_POS))
	{
		dev.seek(oldPos);
		return false;
	}
	stream << (qint32)dataSize;
	
	// Restore the old position.
	dev.seek(oldPos);
	
	return true;
}

// Wave format description
// https://ccrma.stanford.edu/courses/422/projects/WaveFormat/
bool WAVContainer::writeWAVHeader(QIODevice& dev, const QAudioFormat& format)
{
	if (!dev.seek(0))
	{
		return false;
	}

	qint32 byteRate = format.sampleRate() 
		* format.channelCount()
		* (format.sampleSize() / 8);
	qint16 blockAlign = format.channelCount()
		* (format.sampleSize() / 8);

	WAVHeader header;
	
	header.size = (quint32)-1;
	header.totalHeaderSize = WAVHeader::HEADER_SIZE;
	header.format.audioFormat = WAVHeader::FormatChunk::FORMAT_PCM;
	header.format.bitsPerSample = format.sampleSize();
	header.format.byteRate = byteRate;
	header.format.blockAlign = blockAlign;
	header.format.numChannels = format.channelCount();
	header.format.sampleRate = format.sampleRate();
	
	
	header.data.size = (quint32)-1;
	
	return header.writeToDevice(dev);
}
