//------------------------------------------------------------------------------
// wavheader.h
//
// Copyright (C) 2011 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#ifndef __WAVHEADER_H__
#define __WAVHEADER_H__

#include <QByteArray>
#include <QIODevice>

/**
 * @brief Simplified RIFF WAVE header with only Format and Data chunks support.
 */
class WAVHeader
{
	public:
		/**
		 * @brief Format of the data.
		 */
		class FormatChunk
		{
			public:
				/**
				 * @brief PCM format value for audioFormat field.
				 */
				static const qint16 FORMAT_PCM = 1;
			
				/**
				 * @brief Size of the chunk. Not required for save operation.
				 */
				quint32 size;
				
				/**
				 * @brief Only PCM is supported here.
				 */
				quint16 audioFormat;
				
				/**
				 * @brief Number of channels (mono/stereo)
				 */
				quint16 numChannels;
				
				/**
				 * @brief Sample rate (8000, 11250, 22500, etc.)
				 */
				quint32 sampleRate;
				
				/**
				 * @brief == sampleRate * numChannels * bitsPerSample / 8
				 */
				quint32 byteRate;
				
				/**
				 * @brief == numChannels * bitsPerSample / 8
				 *
				 * Number of bytes for one sample with on all channels.
				 */
				quint16 blockAlign;
				
				/**
				 * @brief 8, 16, 24, 32.
				 */
				quint16 bitsPerSample;
								
				/**
				 * @brief Only if audioFormat is other than FORMAT_PCM.
				 */
				QByteArray extraParams;
				
				FormatChunk();
		};
		
		/**
		 * @brief Data information. Doesn't contain actual data.
		 */
		class DataChunk
		{
			public:
				/**
				 * @brief Size of the data.
				 */
				quint32 size;
				
				DataChunk();
		};
		
		static const int DATACHUNK_SIZE_POS = 40;
		static const int HEADER_SIZE_CONSTANT_FOR_MAIN_CHUNK = 36;
		static const int HEADER_SIZE = 44;
		
		/**
		 * @brief Size of data + size of header - 8.
		 */
		quint32 size;
		
		/**
		 * @brief Complete size of the header.
		 *
		 * Unused when writing.
		 */
		quint32 totalHeaderSize;
		
		/**
		 * @brief Format chunk. Contains information about data format.
		 */
		FormatChunk format;
		
		/**
		 * @brief Data chunk. Contains the actual data.
		 *
		 * @b Note:
		 * Data is not included in this structure.
		 */
		DataChunk data;

		/**
		 * @brief Reads the RIFF WAVE header from IO device.
		 *
		 * @b Note:
		 * This method is currently not implemented. It will be implemented
		 * if it becomes necessary to read this information.
		 */
		static WAVHeader readFromDevice(QIODevice& dev);

		WAVHeader();
		~WAVHeader();
		
		bool isValid() const;
	
		/**
		 * @brief Writes this RIFF WAVE header to the IO device.
		 *
		 * The header is written at the position the device is currently
		 * in.
		 */	
		bool writeToDevice(QIODevice& dev);
		
	private:

};

#endif