//------------------------------------------------------------------------------
// filteredaudioinput.cpp
//
// Copyright (C) 2013 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "filteredaudioinput.h"

#include "audio/filters/audiofilter.h"
#include "audio/filters/filterfactory.h"
#include "audio/filters/filtermetadata.h"
#include "audio/filters/filterpipeline.h"
#include "audio/audiocore.h"
#include "audio/audiodevformat.h"
#include "audio/audioerror.h"
#include "audio/audioinputformatfactory.h"
#include "config/audioinputconfig.h"
#include "config/filtersconfig.h"
#include <QAudioInput>
#include <QDebug>

class FilteredAudioInput::PrivData
{
	public:
		QAudioInput* audioInput;
		AudioDevFormat devFormat;
		FilterPipeline* filterPipeline;
		int lastError;
};
////////////////////////////////////////////////////////////////////////////////
FilteredAudioInput::FilteredAudioInput()
{
	d = new PrivData();
	d->audioInput = NULL;
	d->filterPipeline = NULL;
	d->lastError = 0;
}

FilteredAudioInput::~FilteredAudioInput()
{
	if (d->filterPipeline != NULL)
	{
		qDeleteAll(d->filterPipeline->filters());
		delete d->filterPipeline;
	}
	if (d->audioInput != NULL)
	{
		delete d->audioInput;
	}
	delete d;
}

void FilteredAudioInput::createFilterPipeline()
{
	FiltersConfig cfgFilters(gpConfig);
	QStringList filters = cfgFilters.enabledFilters();
	d->filterPipeline = new FilterPipeline();
	this->connect(d->filterPipeline, SIGNAL(readyRead()), SIGNAL(readyRead()));
	FilterMetaData filterMetaData;
	filterMetaData.audioFormat = d->audioInput->format();
	foreach (const QString& filterName, filters)
	{
		AudioFilter* filter = FilterFactory::construct(filterName,
				filterMetaData);
		qDebug() << "Adding filter: " << filterName << filter;
		if (filter != NULL)
		{
			d->filterPipeline->addFilter(filter);
		}
	}
	d->filterPipeline->open(QIODevice::ReadWrite);
}

const AudioDevFormat& FilteredAudioInput::devFormat() const
{
	return d->devFormat;
}

bool FilteredAudioInput::isRecording() const
{
	if (d->audioInput == NULL)
	{
		return false;
	}

	return d->audioInput->state() == QAudio::ActiveState
	    || d->audioInput->state() == QAudio::IdleState;
}

bool FilteredAudioInput::isSetup() const
{
	return d->audioInput != NULL && d->devFormat.isValid();
}

AudioDevFormat FilteredAudioInput::obtainAudioDevFormat() const
{
	AudioInputConfig cfg(gpConfig);
	AudioInputFormatFactory formatFactory;
	formatFactory.setConfig(&cfg);
	QAudioDeviceInfo audioDev = formatFactory.getAudioDeviceInfo();
	if (audioDev.isNull())
	{
		d->lastError = AudioError::DeviceNotFound;
		return AudioDevFormat();
	}
	return AudioDevFormat(audioDev, formatFactory.getAudioFormat(audioDev));
}

QByteArray FilteredAudioInput::readPending() const
{
	return d->filterPipeline->readAll();
}

qint64 FilteredAudioInput::recordingTimeMs() const
{
	qint64 elapsed = d->audioInput->elapsedUSecs() / 1000ULL;
	return elapsed;
}

QAudioInput* FilteredAudioInput::setupAudioInput()
{
	const int ONE_SECOND = 1000;

	if (!AudioCore::hasAnyAudioDevices(QAudio::AudioInput))
	{
		d->lastError = AudioError::NoAudioDevices;
		return NULL;
	}

	AudioDevFormat devFormat = obtainAudioDevFormat();
	if (!devFormat.isValid())
	{
		return NULL;
	}
	QAudioFormat format = devFormat.format();
	qDebug() << QString("Recording format: codec=%1, byteOrder=%2, "
			"channels=%3, sampleRate=%4, sampleSize=%5, sampleType=%6")
		.arg(format.codec())
		.arg(format.byteOrder())
		.arg(format.channelCount())
		.arg(format.sampleRate())
		.arg(format.sampleSize())
		.arg(format.sampleType());

	QAudioInput* pAudioInput = new QAudioInput(devFormat.audioDev(), format, this);
	pAudioInput->setNotifyInterval(ONE_SECOND);
	this->connect(pAudioInput, SIGNAL(notify()), SIGNAL(notify()));
	this->connect(pAudioInput, SIGNAL(stateChanged(QAudio::State)),
		SIGNAL(stateChanged(QAudio::State)));
	d->devFormat = AudioDevFormat(devFormat.audioDev(), pAudioInput->format());

	return pAudioInput;
}

int FilteredAudioInput::setup()
{
	if (isRecording())
	{
		// This method is public.
		// Do not allow to start recording twice.
		return AudioError::OperationAlreadyInProgress;
	}
	d->lastError = 0;

	if (d->audioInput != NULL)
	{
		delete d->audioInput;
		d->audioInput = NULL;
	}
	if (d->filterPipeline != NULL)
	{
		qDeleteAll(d->filterPipeline->filters());
		delete d->filterPipeline;
		d->filterPipeline = NULL;
	}

	d->audioInput = setupAudioInput();
	if (d->audioInput == NULL)
	{
		return d->lastError;
	}

	createFilterPipeline();

	return 0;
}

int FilteredAudioInput::start()
{
	if (isRecording())
	{
		qDebug() << "Attempted to start recording twice.";
		return AudioError::OperationAlreadyInProgress;
	}
	if (!isSetup())
	{
		qDebug() << "Attempted to start recording on an unset recorder.";
		return AudioError::NotSet;
	}
	d->audioInput->start(d->filterPipeline);
	return 0;
}

void FilteredAudioInput::stop()
{
	if (isRecording())
	{
		// This method is a public slot.
		// Do not allow to stop recording if it is not running.
		d->audioInput->stop();
		d->filterPipeline->close();
	}
}
