//------------------------------------------------------------------------------
// audiocfgtab.cpp
//
// Copyright (C) 2012 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "gui/configtabs/audiocfgtab.h"

#include "audio/audiocore.h"
#include "audio/audioinputformatfactory.h"
#include "config/audioinputconfig.h"

#include <QDebug>
#include <QMetaEnum>
#include <QMetaObject>

AudioCfgTab::AudioCfgTab(QWidget* pParent)
: ConfigTab(pParent)
{
	setupUi(this);
}

void AudioCfgTab::audioDeviceChanged(int index)
{
	fillFormatComboBoxes();
	QAudioFormat format = currentAudioFormat();

	if (format.isValid())
	{
		QAudioDeviceInfo dev = currentAudioDevice();
		setAudioFormat(dev.nearestFormat(format));
	}
}

QAudioDeviceInfo AudioCfgTab::currentAudioDevice() const
{
	return AudioInputFormatFactory::getAudioDeviceInfo(cboAudioInputSource->currentText());
}

QAudioFormat AudioCfgTab::currentAudioFormat() const
{
	QAudioFormat format;

	format.setCodec(AudioCore::AUDIO_PCM_CODEC);
	format.setByteOrder((QAudioFormat::Endian)cboByteOrder->itemData(
		cboByteOrder->currentIndex()).toInt());

	format.setChannelCount(cboChannels->itemData(cboChannels->currentIndex()).toInt());
	format.setSampleRate(cboSampleRate->itemData(cboSampleRate->currentIndex()).toInt());
	format.setSampleSize(cboSampleSize->itemData(cboSampleSize->currentIndex()).toInt());
	format.setSampleType((QAudioFormat::SampleType)cboSampleType->itemData(
		cboSampleType->currentIndex()).toInt());

	return format;
}

void AudioCfgTab::fillComboBox(QComboBox* pComboBox, const QList<int>& values)
{
	pComboBox->clear();
	foreach (int val, values)
	{
		pComboBox->addItem(QString("%1").arg(val), val);
	}
}

void AudioCfgTab::fillFormatComboBoxes()
{
	QAudioDeviceInfo dev = currentAudioDevice();

	QList<QAudioFormat::Endian> byteOrders = dev.supportedByteOrders();
	QList<int> channels = dev.supportedChannelCounts();
	QList<int> sampleRates = dev.supportedSampleRates();
	QList<int> sampleSizes = dev.supportedSampleSizes();
	QList<QAudioFormat::SampleType> sampleTypes = dev.supportedSampleTypes();

	fillComboBox(cboChannels, channels);
	fillComboBox(cboSampleRate, sampleRates);
	fillComboBox(cboSampleSize, sampleSizes);

	// Byte order
	cboByteOrder->clear();
	if (byteOrders.contains(QAudioFormat::LittleEndian))
	{
		cboByteOrder->addItem(AudioCore::audioFormatEndianEnumToString(QAudioFormat::LittleEndian),
			QAudioFormat::LittleEndian);
	}

	if (byteOrders.contains(QAudioFormat::BigEndian))
	{
		cboByteOrder->addItem(AudioCore::audioFormatEndianEnumToString(QAudioFormat::BigEndian),
			QAudioFormat::BigEndian);
	}

	// Sample Types
	cboSampleType->clear();
	if (sampleTypes.contains(QAudioFormat::SignedInt))
	{
		cboSampleType->addItem(
			AudioCore::audioFormatSampleTypeEnumToString(QAudioFormat::SignedInt),
			QAudioFormat::SignedInt);
	}

	if (sampleTypes.contains(QAudioFormat::UnSignedInt))
	{
		cboSampleType->addItem(
			AudioCore::audioFormatSampleTypeEnumToString(QAudioFormat::UnSignedInt),
			QAudioFormat::UnSignedInt);
	}

	if (sampleTypes.contains(QAudioFormat::Float))
	{
		cboSampleType->addItem(
			AudioCore::audioFormatSampleTypeEnumToString(QAudioFormat::Float),
			QAudioFormat::Float);
	}
}

void AudioCfgTab::hideErrorLabels()
{
}

void AudioCfgTab::loadAudioInputDevices()
{
	AudioInputConfig cfg(gpConfig);
	cboAudioInputSource->clear();

	// We need to call this first, otherwise the retrieved list might not be
	// complete.
	QAudioDeviceInfo::defaultInputDevice();

	// Retrieve all known devices and load them into ComboBox.
	QList<QAudioDeviceInfo> inputDevices = QAudioDeviceInfo::availableDevices(QAudio::AudioInput);
	foreach (const QAudioDeviceInfo& dev, inputDevices)
	{
		QString name = dev.deviceName();
		qDebug() << "Detected audio input device:" << name;
		cboAudioInputSource->addItem(name);
	}

	// Select the default device in ComboBox.
	if (!selectAudioInputDevice(cfg.deviceName()))
	{
		selectDefaultAudioInputDevice();
	}
}

void AudioCfgTab::loadConfig()
{
	loadAudioInputDevices();
	fillFormatComboBoxes();
	setupFormatParameters();
}

void AudioCfgTab::saveConfig()
{
	AudioInputConfig audioCfg(gpConfig);

	QAudioFormat format = currentAudioFormat();

	audioCfg.setDeviceName(cboAudioInputSource->currentText());
	audioCfg.setByteOrder(format.byteOrder());
	audioCfg.setChannelCount(format.channelCount());
	audioCfg.setSampleRate(format.sampleRate());
	audioCfg.setSampleSize(format.sampleSize());
	audioCfg.setSampleType(format.sampleType());
}

bool AudioCfgTab::selectAudioInputDevice(const QString& name)
{
	for (int i = 0; i < cboAudioInputSource->count(); ++i)
	{
		const QString& text = cboAudioInputSource->itemText(i);
		if (text.compare(name, Qt::CaseInsensitive) == 0)
		{
			cboAudioInputSource->setCurrentIndex(i);
			return true;
		}
	}

	return false;
}

void AudioCfgTab::selectDefaultAudioInputDevice()
{
	QAudioDeviceInfo defaultDev = QAudioDeviceInfo::defaultInputDevice();

	qDebug() << "Default audio input device is:" << defaultDev.deviceName();

	selectAudioInputDevice(defaultDev.deviceName());
}

void AudioCfgTab::setAudioFormat(const QAudioFormat& format)
{
	cboByteOrder->setCurrentIndex(cboByteOrder->findData(format.byteOrder()));
	cboChannels->setCurrentIndex(cboChannels->findData(format.channelCount()));
	cboSampleRate->setCurrentIndex(cboSampleRate->findData(format.sampleRate()));
	cboSampleSize->setCurrentIndex(cboSampleSize->findData(format.sampleSize()));
	cboSampleType->setCurrentIndex(cboSampleType->findData(format.sampleType()));
}

void AudioCfgTab::setupFormatParameters()
{
	AudioInputConfig cfg(gpConfig);
	AudioInputFormatFactory factory;

	QAudioDeviceInfo dev = currentAudioDevice();

	factory.setConfig(&cfg);
	QAudioFormat format = factory.getAudioFormat(dev);

	if (!format.isValid())
	{
		format = dev.preferredFormat();
	}

	setAudioFormat(format);
}

QString AudioCfgTab::validate()
{
	hideErrorLabels();

	return QString();
}
