//------------------------------------------------------------------------------
// mainwindow.cpp
//
// Copyright (C) 2011 - 2012 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "mainwindow.h"
#include "versioninfo.h"

#include <QDebug>
#include <QDesktopWidget>
#include <QDesktopServices>
#include <QDir>
#include <QFileInfo>
#include <QInputDialog>
#include <QMessageBox>
#include <QTime>
#include <QTimer>
#include <QUrl>

#include "audio/audiocore.h"
#include "config/generalconfig.h"
#include "config/notificationsconfig.h"
#include "config/shortcutsconfig.h"
#include "gui/aboutdialog.h"
#include "gui/configdialog.h"
#include "gui/trayicon.h"
#include "keyboard/keydecoder.h"
#include "keyboard/keystatemonitor.h"
#include "keyboard/shortcut.h"
#include "keyboard/shortcutsmonitor.h"
#include "application.h"

MainWindow* MainWindow::pInstance = NULL;

MainWindow::MainWindow(QWidget* pParent)
: QMainWindow(pParent)
{
	setupUi(this);
	doConnections();

	setTitleToDefault();

	centerOnDesktop();

	bForceClosing = false;
	bWasShown = false;
	bWasMaximized = this->isMaximized();
	pTrayIcon = NULL;
	pTrayIconMenu = NULL;

	setupShortcuts();
	setTrayIconVisibilityAccordingToConfig();
}

MainWindow::~MainWindow()
{

}

void MainWindow::adjustHeight()
{
	// Sum height of margins from top and bottom of the available
	// screen.
	const static int MARGIN = 50 * 2;

	QDesktopWidget* pDesktop = QApplication::desktop();

	if (pDesktop != NULL)
	{
		// Get screen where MainWindow currently resides.
		QRect rectDesktop = pDesktop->availableGeometry(this);

		int hDesktop = rectDesktop.height();

		this->resize(this->width(), hDesktop - MARGIN);
	}
}

void MainWindow::applyShortcutsConfiguration()
{
	ShortcutsConfig cfg(gpConfig);

	qDebug() << "Start recording set to: " << cfg.startRecord();
	pShortcutStartRecording->setKeys(KeyDecoder::parseSequence(cfg.startRecord()));

	qDebug() << "Stop recording set to: " << cfg.stopRecord();
	pShortcutStopRecording->setKeys(KeyDecoder::parseSequence(cfg.stopRecord()));

	qDebug() << "Toggle recording set to: " << cfg.toggleRecord();
	pShortcutToggleRecording->setKeys(KeyDecoder::parseSequence(cfg.toggleRecord()));
}

void MainWindow::bringUp()
{
	if (isMinimized() || !isVisible())
	{
		if (wasMaximized())
		{
			showMaximized();
		}
		else
		{
			showNormal();
		}
	}

	raise();
	activateWindow();
}

void MainWindow::browseOutputDir()
{
	const QString ERROR_CAPTION =  tr("Open output directory");

	GeneralConfig cfg(gpConfig);
	QFileInfo fi(cfg.outputFile());
	QString dirPath = fi.absolutePath();
	fi = QFileInfo(dirPath);
	
	if (!fi.exists())
	{
		QMessageBox::warning(this, ERROR_CAPTION,
			tr("Cannot open directory \"%1\" because it does not exist").arg(dirPath));
		return;
	}
	
	if (!fi.isDir())
	{
		QMessageBox::warning(this, ERROR_CAPTION,
			tr("Failed to open directory in path: \"%1\"\nThis is not a directory.").arg(dirPath));
		return;
	}
	
	dirPath = QDir::toNativeSeparators(dirPath);
	QDesktopServices::openUrl(QUrl("file:///" + dirPath));
}

bool MainWindow::canCloseToTrayIcon() const
{
	GeneralConfig cfg(gpConfig);	

	return cfg.isCloseToTrayIconEnabled()
		&& isTrayIconVisible();
}

bool MainWindow::canMinimizeToTrayIcon() const
{
	GeneralConfig cfg(gpConfig);	

	return cfg.isMinimizeToTrayIconEnabled()
		&& isTrayIconVisible();
}

void MainWindow::centerOnDesktop()
{
	QDesktopWidget* pDesktop = QApplication::desktop();

	if (pDesktop != NULL)
	{
		// Get screen where MainWindow currently resides.
		QRect rectDesktop = pDesktop->availableGeometry(this);

		int wDesktop = rectDesktop.width();
		int hDesktop = rectDesktop.height();

		int wWindow = this->width();
		int hWindow = this->height();

		int x = (wDesktop - wWindow) / 2;
		int y = (hDesktop - hWindow) / 2;

		this->move(x, y);
	}

}

void MainWindow::closeForce()
{
	bForceClosing = true;
	close();	
}

void MainWindow::doConnections()
{
}

bool MainWindow::event(QEvent* pEvent)
{
	if (pEvent->type() == QEvent::WindowStateChange)
	{
		if (this->isMinimized())
		{
			if (canMinimizeToTrayIcon())
			{
				// Stupid solution to a stupid Qt problem.
				// If we call hide() immediately a fake Task Bar
				// entry will be spawned. To avoid this, we use
				// a timer!
				QTimer::singleShot(1, this, SLOT( hide() ) );
			}
		}
		else
		{
			bWasMaximized = this->isMaximized();
		}
	}
	else if (pEvent->type() == QEvent::Close)
	{
		if (canCloseToTrayIcon() && !bForceClosing)
		{
			QTimer::singleShot(1, this, SLOT( hide() ) );
			pEvent->ignore();
			return true;
		}
	}
	else if (pEvent->type() == QEvent::Show)
	{
		if (!bWasShown)
		{
			bWasShown = true;

			// Make sure that the method is run after the program starts
			// the main loop.
			QTimer::singleShot(1, this, SLOT( firstShow() ));
		}
	}

	return QMainWindow::event(pEvent);
}

void MainWindow::firstShow()
{
	if (gpConfig->isFirstRun())
	{
		QMessageBox::information(this, tr("%1 - First Run").arg(VersionInfo::fullName()),
			tr("This is the first run of the application. Please take your time to configure it properly."));

		showConfigurationDialog();
	}
}

void MainWindow::keyStateChanged(KeyInfo::KeyId key, int state)
{
	#ifdef DUMP_ALL_KEYS_TO_STDOUT
		#ifdef NDEBUG
			#error "Key dumping cannot be enabled in release builds \
because this creates a security vulnerability."
		#endif
	qDebug() << "key state changed - keyId:" << key
		<< "->" << KeyInfo::fromKeyId(key).codeName()
		<< "[" << KeyInfo::fromKeyId(key).keyId() << "], new state:" << state;
	#endif
}

bool MainWindow::isTrayIconVisible() const
{
	if (pTrayIcon == NULL)
	{
		return false;
	}

	return pTrayIcon->isVisible();
}

void MainWindow::recordingNotify()
{
	NotificationsConfig cfg(gpConfig);

	if (cfg.isShowRecordingTimeOnTitleBarEnabled())
	{
		// TODO: Refactor - String building should happen in a different class.
		qint64 msecs = recordArea->recordingTimeMs();
		qint64 secs = msecs / 1000;
		qint64 mins = secs / 60;
		qint64 hours = mins / 60;
		
		secs %= 60;
		mins %= 60;

		QString strTime;
		if (hours > 0)
		{	
			strTime = QString("%1:%2:%3")
				.arg(hours, 2, 10, QLatin1Char('0'))
				.arg(mins, 2, 10, QLatin1Char('0'))
				.arg(secs, 2, 10, QLatin1Char('0'));
		}
		else
		{
			strTime = QString("%1:%2")
				.arg(mins, 2, 10, QLatin1Char('0'))
				.arg(secs, 2, 10, QLatin1Char('0'));
		}
			
		this->setWindowTitle(QString("[%1] %2")
			.arg(strTime).arg(VersionInfo::fullName()));
	}
}

void MainWindow::recordingStateChanged(bool bIsRecording)
{
	setTrayIconAppearanceAccordingToState();
	if (!recordArea->newestRecordedFilePath().isEmpty())
	{
		if (recordArea->isRecording())
		{
			lblLastRecordedFile->setText(tr("Current recording:"));
		}
		else
		{
			lblLastRecordedFile->setText(tr("Last recording:"));
		}
		leLastRecordedFile->setText(recordArea->newestRecordedFilePath());
	}
	
	if (recordArea->isRecording())
	{
		const int BITS_IN_ONE_BYTE = 8;
	
		widgetRecordingInfo->show();
		
		const QAudioFormat& format = recordArea->newestRecordingFormat();
		
		QString sampleType = AudioCore::audioFormatSampleTypeEnumToString(format.sampleType());
		QString byteOrder;
		
		if (format.sampleSize() > BITS_IN_ONE_BYTE)
		{
			byteOrder = AudioCore::audioFormatEndianEnumToString(format.byteOrder());
		}
		else
		{
			byteOrder = "N/A";
		}
		
		lblChannels->setText(QString("%1").arg(format.channelCount()));
		lblSampleRate->setText(QString("%1").arg(format.sampleRate()));
		lblSampleType->setText(sampleType);
		lblSampleSize->setText(QString("%1").arg(format.sampleSize()));
		lblByteOrder->setText(byteOrder);
	}
	else
	{
		setTitleToDefault();
	}
}

void MainWindow::setTitleToDefault()
{
	this->setWindowTitle(VersionInfo::fullName());
}

void MainWindow::setTrayIconAppearanceAccordingToState()
{
	if (isTrayIconVisible())
	{
		if (recordArea->isRecording())
		{
			pTrayIcon->setIconRecording();
		}
		else
		{
			pTrayIcon->setIconNormal();
		}
	}
}

void MainWindow::setTrayIconVisibilityAccordingToConfig()
{
	GeneralConfig cfg(gpConfig);

	if (pTrayIcon != NULL)
	{
		pTrayIcon->setVisible(cfg.isTrayIconEnabled());
	}
	else if (cfg.isTrayIconEnabled())
	{
		setupTrayIcon();
	}
	
	// This will also perform check to see if tray icon is present.
	setTrayIconAppearanceAccordingToState();
}

void MainWindow::setupShortcuts()
{
	this->connect(gpApp->keyStateMonitor(),
		SIGNAL( keyStateChanged(KeyInfo::KeyId, int) ),
		SLOT  ( keyStateChanged(KeyInfo::KeyId, int) ) );

	pShortcutStartRecording = new Shortcut();
	pShortcutStopRecording = new Shortcut();
	pShortcutToggleRecording = new Shortcut();

	connect(pShortcutStartRecording,
		SIGNAL( shortcutFired() ),
		recordArea,
		SLOT  ( startRecording() ) );

	connect(pShortcutStopRecording,
		SIGNAL( shortcutFired() ),
		recordArea,
		SLOT  ( stopRecording() ) );

	connect(pShortcutToggleRecording,
		SIGNAL( shortcutFired() ),
		recordArea,
		SLOT  ( toggleRecording() ) );

	gpApp->registerShortcut(pShortcutStartRecording);
	gpApp->registerShortcut(pShortcutStopRecording);
	gpApp->registerShortcut(pShortcutToggleRecording);

	applyShortcutsConfiguration();
}

void MainWindow::setupTrayIcon()
{
	pTrayIcon = new TrayIcon(this);
	
	pTrayIconMenu = new TrayIconMenu(this);
	this->connect(pTrayIconMenu,
		SIGNAL( triggered(QAction*) ),
		SLOT( trayIconMenuActivated(QAction*) ) );
	
	pTrayIcon->setContextMenu(pTrayIconMenu);
	pTrayIcon->show();
}

void MainWindow::showAboutDialog()
{
	gpApp->shortcutsMonitor()->setEnabled(false);
	AboutDialog(this).exec();
	gpApp->shortcutsMonitor()->setEnabled(true);
}

void MainWindow::showAboutQtDialog()
{
	gpApp->shortcutsMonitor()->setEnabled(false);
	QApplication::aboutQt();
	gpApp->shortcutsMonitor()->setEnabled(true);
}

void MainWindow::showConfigurationDialog()
{
	// We need to make sure that keystrokes are not detected while we
	// are setting them!
	gpApp->shortcutsMonitor()->setEnabled(false);

	ConfigDialog dialog(this);
	if (dialog.exec() == QDialog::Accepted)
	{
		applyShortcutsConfiguration();
		setTrayIconVisibilityAccordingToConfig();
		
		{
			NotificationsConfig cfg(gpConfig);
			if (!cfg.isShowRecordingTimeOnTitleBarEnabled())
			{
				// If we disabled the time display then it's a good idea
				// to update the title instantly. If the above settings didn't
				// change then the following operation will do nothing.
				setTitleToDefault();
			}
		}
	}
	gpApp->shortcutsMonitor()->setEnabled(true);
}

void MainWindow::trayIconMenuActivated(QAction* pAction)
{
	if (pAction == NULL)
	{
		return;
	}

	if (pAction == pTrayIconMenu->pActionShow)
	{
		this->bringUp();
	}
	else if (pAction == pTrayIconMenu->pActionExit)
	{
		this->closeForce();
	}
	else
	{
		QMessageBox::warning(this, tr("Tray Icon Menu"),
			tr("Unsupported operation."));
	}
}

//////////////////////////////////////////////////////////////////////////////
MainWindow::TrayIconMenu::TrayIconMenu(QWidget* pParent)
: QMenu(pParent)
{
	pActionShow = new QAction(QIcon(":/icons/zoom"), tr("Show"), this);
	pActionExit = new QAction(QIcon(":/icons/remove"), tr("Exit"), this);

	this->addAction(pActionShow);
	this->addSeparator();
	this->addAction(pActionExit);
}

MainWindow::TrayIconMenu::~TrayIconMenu()
{
	qDebug() << "~TrayIconMenu()";
}
