//------------------------------------------------------------------------------
// recordarea.cpp
//
// Copyright (C) 2011 - 2012 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "recordarea.h"

#include "audio/audioerror.h"
#include "audio/audiorecorder.h"
#include "config/generalconfig.h"
#include "config/notificationsconfig.h"
#include "paths.h"
#include <QDebug>
#include <QFile>
#include <QMessageBox>
#include <QSound>

class RecordArea::PrivData
{
	public:
		AudioRecorder* recorder;
		QString newestRecordedFile;
		QFile* outputFile;
};
///////////////////////////////////////////////////////////////////////////////
RecordArea::RecordArea(QWidget* pParent)
: QWidget(pParent)
{
	setupUi(this);
	d = new PrivData();

	lblTime->setTime(0);
	d->recorder = NULL;
	d->outputFile = NULL;
}

RecordArea::~RecordArea()
{
	if (d->recorder != NULL)
	{
		delete d->recorder;
	}
	if (d->outputFile != NULL)
	{
		delete d->outputFile;
	}
	delete d;
}

void RecordArea::audioInputNotify()
{
	qint64 elapsed = recordingTimeMs();
	lblTime->setTime(elapsed);
	emit recordingNotify();
}

void RecordArea::audioInputStateChanged(QAudio::State state)
{
	switch (state)
	{
		case QAudio::ActiveState:
			qDebug() << "QAudio::ActiveState";
			switchStateToRecording();
			break;

		case QAudio::SuspendedState:
			qDebug() << "QAudio::SuspendedState";
			break;

		case QAudio::StoppedState:
			qDebug() << "QAudio::StoppedState";
			switchStateToStopped();
			break;

		case QAudio::IdleState:
			qDebug() << "QAudio::IdleState";
			break;

		default:
			qDebug() << "Unhandled QAudio::State in RecordArea::recordingStateChanged(): "
				<< state;
			break;
	}
}

QString RecordArea::generateOutputPath()
{
	GeneralConfig config(gpConfig);
	QString outPath = config.outputFile();
	outPath = Paths::resolveOutputFilePath(outPath);
	outPath = Paths::ensureUniquePath(outPath);
	return outPath;
}

bool RecordArea::isRecording() const
{
	if (d->recorder == NULL)
	{
		return false;
	}
	return d->recorder->isRecording();
}

const QString& RecordArea::newestRecordedFilePath() const
{
	return d->newestRecordedFile;
}

const QAudioFormat& RecordArea::newestRecordingFormat() const
{
	return d->recorder->recordingFormat();
}

qint64 RecordArea::recordingTimeMs() const
{
	return d->recorder->recordingTimeMs();
}

void RecordArea::setupRecorder()
{
	d->recorder = new AudioRecorder();
	this->connect(d->recorder, SIGNAL( stateChanged(QAudio::State) ),
		SLOT( audioInputStateChanged(QAudio::State) ) );
	this->connect(d->recorder, SIGNAL( notify() ),
		SLOT( audioInputNotify() ) );
}

void RecordArea::startRecording()
{
	if (isRecording())
	{
		// This method is a public slot.
		// Do not allow to start recording twice.
		return;
	}

	if (d->recorder != NULL)
	{
		delete d->recorder;
	}
	if (d->outputFile != NULL)
	{
		delete d->outputFile;
	}

	setupRecorder();
	
	QString outputFilePath = generateOutputPath();
	d->outputFile = new QFile(outputFilePath);
	if (!d->outputFile->open(QIODevice::WriteOnly | QIODevice::Truncate))
	{
		QMessageBox::critical(this, tr("Cannot start recording"),
			tr("Cannot open file %1 for writing.").arg(outputFilePath));
	}
	int errorCode = d->recorder->startRecording(d->outputFile);
	if (errorCode != 0)
	{
		QMessageBox::critical(this, tr("Cannot start recording"),
			AudioError::toString(errorCode));
	}
	d->newestRecordedFile = outputFilePath;
}

void RecordArea::stopRecording()
{
	if (isRecording())
	{
		// This method is a public slot.
		// Do not allow to stop recording if it is not running.
		d->recorder->stopRecording();
		d->outputFile->close();
	}
}

void RecordArea::switchStateToRecording()
{
	btnRecord->setEnabled(false);
	btnStop->setEnabled(true);

	NotificationsConfig cfg(gpConfig);
	if (cfg.isStartedRecordingSoundEnabled())
	{
		QSound::play(cfg.startedRecordingSoundPath());
	}

	emit recordingStateChanged(true);
}

void RecordArea::switchStateToStopped()
{
	btnRecord->setEnabled(true);
	btnStop->setEnabled(false);
	
	NotificationsConfig cfg(gpConfig);
	if (cfg.isStoppedRecordingSoundEnabled())
	{
		QSound::play(cfg.stoppedRecordingSoundPath());
	}
	emit recordingStateChanged(false);
}

void RecordArea::toggleRecording()
{
	if (isRecording())
	{
		stopRecording();
	}
	else
	{
		startRecording();
	}
}
