//------------------------------------------------------------------------------
// keystatemonitor.h
//
// Copyright (C) 2011 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#ifndef __KEYSTATEMONITOR_H__
#define __KEYSTATEMONITOR_H__

#include "keyboard/keyinfo.h"

#include <QMap>
#include <QTimer>
#include <QSet>

class KeyStateReader;

class KeyStateMonitor : public QObject
{
	Q_OBJECT

	public:
		enum KeyState
		{
			Depressed,
			Pressed
		};

		KeyStateMonitor();
		
		/**
		 * @brief Gets all keys that have KeyState::Pressed state.
		 */
		QList<KeyInfo::KeyId> pressedKeys() const;
		
		/**
		 * @brief Resumes monitoring after stop.
		 *
		 * This works only if the following sequence of methods is called:
		 * @code
		 * KeyStateMonitor monitor;
		 * monitor.startMonitoring(pKeyStateReader, 20);
		 * monitor.stopMonitoring();
		 * monitor.resumeMonitoring();
		 * @endcode
		 *
		 * @return <code>true</code> if monitoring was resumed correctly.
		 */
		bool resumeMonitoring();

		/**
		 * @brief Sets all keys that should be monitored.
		 */
		void setKeys(QSet<KeyInfo::KeyId> keys);
		
		/**
		 * @brief Starts time based monitoring of keys.
		 *
		 * @param pKeyStateReader
		 *      Keyboard state reader. The monitor will @b not take ownership
		 *      of this object and will @b not delete it when being deleted 
		 *      itself.
		 * @param intervalMs
		 *      Interval at which the keyboard state will be read.
		 */
		void startMonitoring(KeyStateReader* pKeyStateReader, unsigned intervalMs);

		/**
		 * @brief Entire current state of all monitored keys.
		 */		
		const QMap<KeyInfo::KeyId, int>& state() const;
		
		/**
		 * @brief Stops time based monitoring of keys.
		 *
		 * @see startMonitoring()
		 */
		void stopMonitoring();
		
		/**
		 * @brief Updates the internal keyboard state map and fires
		 *        keyStateChanged() signals.
		 *
		 * keyStateChanged() signal is fired for each key for which the
		 * state changed. States are defined by KeyState enum.
		 */
		void updateStates(const QMap<KeyInfo::KeyId, int>& newKeyStates);

	signals:
		/**
		 * @brief Emitted for all keys that changed in a single monitor
		 *        loop run.
		 *
		 * @param keyStates
		 *      A QMap where keys are of type KeyInfo::KeyId and values are
		 *      of type KeyState.
		 */
		void keyBatchStateChanged(QMap<KeyInfo::KeyId, int> keyStates);
		
		/**
		 * @brief Emitted for a single key for which the state has changed.
		 *
		 * @param key
		 *      KeyInfo::KeyId
		 * @param newState
		 *      KeyState
		 */
		void keyStateChanged(KeyInfo::KeyId key, int newState);

	private:
		QMap<KeyInfo::KeyId, int> keyStateMap;
		QTimer monitorTimer;
		KeyStateReader* pKeyStateReader;
		
	private slots:
		void readKeyboardState();
};

#endif
