//------------------------------------------------------------------------------
// main.cpp
//
// Copyright (C) 2011 "Zalewa" <zalewapl@gmail.com>
//------------------------------------------------------------------------------
#include "main.h"

#include <QDebug>
#include <QDir>
#include <QMessageBox>
#include <QString>
#include <QStringList>

#include "audio/audiocore.h"
#include "audio/audioerror.h"
#include "config/config.h"
#include "gui/mainwindow.h"
#include "application.h"
#include "main.h"
#include "paths.h"
#include "versioninfo.h"

Main* Main::pMain = NULL;

Main::Main(int argc, char** argv)
{
	pMain = this;

	this->argc = argc;
	this->argv = argv;
}

Main::~Main()
{
}

int Main::run()
{
	// Pre-init setup.
	application = new Application(argc, argv);
	qDebug() << "==================================";
	qDebug() << "Starting " << VersionInfo::fullName() << ". Hello World :)";

	// Init Qt
	qDebug() << "Setting QCoreApplication info";
	QCoreApplication::setApplicationName(VersionInfo::programName());
	QCoreApplication::setApplicationVersion(VersionInfo::versionRevision());

	// Create local data directory.
	qDebug() << "Creating directory: " << Paths::documentsPath();
	QDir documentsDir(Paths::documentsPath());
	if (!documentsDir.mkpath("."))
	{
		qDebug() << "Failed to spawn path: " << documentsDir.absolutePath()
			<< ", but the application will continue.";
	}

	// Init config
	qDebug() << "Setting up config";
	Config::init("ari.ini");
	gpConfig->synchronize();

	qDebug() << "Config file path is: " << gpConfig->settings().fileName();

	// Check if audio input is supported by the OS
	if (!AudioCore::hasAnyAudioDevices(QAudio::AudioInput))
	{
		QString caption = QObject::tr("%1 - Audio Input error").arg(VersionInfo::programName());
		QString error = QObject::tr("<b>Critical error:</b><br><br>");
		error += AudioError::toString(AudioError::NoAudioDevices);
		error += QObject::tr("<br>The application will now start but it will not work correctly.");

		QMessageBox::critical(NULL, caption, error);
	}

	// First run procedure
	if (gpConfig->isFirstRun())
	{
		firstRun();
	}

	// Init main window
	MainWindow::init();

	// Launch application.
	int returnValue = 0;
	gpMainWindow->show();

	qDebug() << "Init finished.";
	qDebug() << "==================================";
	returnValue = application->exec();

	// Post run clean-up.
	qDebug() << "==================================";
	qDebug() << "Closing application. Post run clean-up.";

	// Singletons clean-up.
	qDebug() << "Destroying main window.";
	MainWindow::destroy();

	// Deinit configuration.
	qDebug() << "Closing configuration.";
	Config::deinit();

	qDebug() << "Bye!";
	qDebug() << "==================================";
	return returnValue;
}

void Main::firstRun()
{
	QMessageBox::information(NULL, QObject::tr("%1 - First Run").arg(VersionInfo::fullName()),
		QObject::tr("<b>Important:</b><br>"
			"Anti-virus software may detect this program as a keylogger.<br><br>"
			"This happens because this application needs to read the state "
			"of your keyboard in regular intervals in order to make the global hotkeys work.<br><br>"
			"This application doesn't record your keystrokes and it doesn't "
			"send any kind of information to any third parties."
		)
	);
	#ifdef Q_WS_X11
	QMessageBox::information(NULL, QObject::tr("%1 - Linux - First Run").arg(VersionInfo::fullName()),
		QObject::tr("<b>Important:</b><br>"
			"Support for Linux is just a secondary addition and not a primary objective.<br>"
			"However, if you detect any error, in particular any error with keyboard definitions, "
			"do not hesitate to file a bug report. "
		)
	);
	#endif
}


////////////////////////////////////////////////////////////////////////////////
// Entry point
////////////////////////////////////////////////////////////////////////////////

#ifdef _MSC_VER
	#ifndef _DEBUG
		#define USE_WINMAIN_AS_ENTRY_POINT
	#endif
#endif

#ifdef USE_WINMAIN_AS_ENTRY_POINT
#include <windows.h>
void getCommandLineArgs(QStringList& outList)
{
	// Conversion to "ushort*" seems to work for LPWSTR.
	QString commandLine = QString::fromUtf16((const ushort*)GetCommandLineW());

	int numArgs = 0;
	LPWSTR* winapiCmdLine = CommandLineToArgvW(GetCommandLineW(), &numArgs);

	if (winapiCmdLine == NULL)
	{
		return;
	}

	for (int i = 0; i < numArgs; ++i)
	{
		outList << QString::fromUtf16((const ushort*)winapiCmdLine[i]);
	}
	LocalFree(winapiCmdLine);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR szCmdLine, int nCmdShow)
{
	int argc = 0;
	char** argv = NULL;

	// Good job Microsoft. Now I have to work around your decision of removing
	// useful argc/argv parameters.
	QStringList commandLine;
	getCommandLineArgs(commandLine);

	// At least one is ensured to be here.
	argc = commandLine.size();
	argv = new char*[argc];

	for (int i = 0; i < commandLine.size(); ++i)
	{
		const QString& parameter = commandLine[i];
		argv[i] = new char[parameter.size() + 1];
		strcpy(argv[i], parameter.toAscii().constData());
	}

	Main* pMain = new Main(argc, argv);
	int returnValue = pMain->run();

	// Cleans up after the program.
	delete pMain;

	// On the other hand we could just ignore the fact that this array is left
	// hanging in the memory because Windows will clean it up for us...
	for (int i = 0; i < argc; ++i)
	{
		delete [] argv[i];
	}
	delete [] argv;


	return returnValue;
}
#else
int main(int argc, char* argv[])
{
	Main* pMain = new Main(argc, argv);
	int returnValue = pMain->run();

	// Cleans up after the program.
	delete pMain;

	return returnValue;
}
#endif
